/*
    K8400 firmware modification based on original Version provided by Vellemann.

 Copyright (C) 2017 Holger Höhnemann

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include "cardreader.h"
#include "printer.h"
#include "language.h"
#include "Service.h"
#include "stepper.h"
#include "temperature.h"

#ifdef SDSUPPORT



CardReader::CardReader()
{
   filesize = 0;
   sdpos = 0;
   saving = false;
   logging = false;
   autostart_atmillis=0;
   current_path[0] = '\0';
   workDirDepth = 0;
   file_subcall_ctr=0;
   //mset(workDirParents, 0, sizeof(workDirParents));

   autostart_stilltocheck=true; //the SD start is delayed, because otherwise the serial cannot answer fast enough to make contact with the host software.
   lastnr=0;
   in_file = out_file = NULL;

  autostart_atmillis=millis()+5000;
}

//char *createFilename(char *buffer,const dir_t &p) //buffer>12characters
//{
//  char *pos=buffer;
//  for (uint8_t i = 0; i < 11; i++)
//  {
//    if (p.name[i] == ' ')continue;
//    if (i == 8)
//    {
//      *pos++='.';
//    }
//    *pos++=p.name[i];
//  }
//  *pos++=0;
//  return buffer;
//}


void  CardReader::LsDive(FatFile *parent, uint8_t flags, uint8_t indent)
{
	dir_t p;
	uint8_t cnt=0;

	FatFile file ;
	parent->rewind();
	while (file.openNext(parent, O_READ))
	{
		// indent for dir level
		if (!file.isHidden() || (flags & LS_A))
		{
			for (uint8_t i = 0; i < indent; i++)
			{
				my_printer.message(PSTR(" "));
			}
			//    if (flags & LS_DATE) {
			//      file.printModifyDateTime(pr);
			//      pr->write(' ');
			//    }
			//    if (flags & LS_SIZE) {
			//      file.printFileSize(pr);
			//      pr->write(' ');
			//    }
			file.getName(my_printer.buffer,70);
			my_printer.message(my_printer.buffer);
			if (file.isDir())
			{
				my_printer.message(PSTR("/ "));
			}
			my_printer.message(PSTR("\r\n"));
			if ((flags & LS_R) && file.isDir())
			{
				LsDive(&file, flags, indent + 2);
			}
		}
		file.close();
	}
}
void CardReader::ls() 
{
sd.chdir();
sd.vwd()->rewind();
LsDive (sd.vwd(),0,1);
}


void CardReader::initsd()
{
	//  for card size
	uint32_t cardSize;
	Status = No_Card;

#ifdef SDSLOW
	if (!card.init(SPI_HALF_SPEED,SDSS))
#else

		if (!sd.begin(SDSS,SPI_FULL_SPEED))
#endif
		{

			if (!sd.begin(SDSS,SPI_HALF_SPEED))

				my_printer.echo_ln(T(MSG_SD_Init_Fail));
		}

	cardSize = sd.card()->cardSize();
	if (cardSize == 0) {
		my_printer.echo_ln(T(MSG_SD_Init_Fail));
		return;
	}
	my_printer.message(PSTR("Card type: "));
	switch (sd.card()->type()) {
	case SD_CARD_TYPE_SD1:
		my_printer.message(PSTR("SD1\n"));
		break;

	case SD_CARD_TYPE_SD2:
		my_printer.message(PSTR("SD2\n"));
		break;

	case SD_CARD_TYPE_SDHC:
		if (cardSize < 70000000) {
			my_printer.message(PSTR("SDHC\n "));

		} else {
			my_printer.message(PSTR("SDXC\n"));
		}
		break;
	default:
		my_printer.message(PSTR("Unknown\n"));
	}
	my_printer.f_message(PSTR("Card size: %iMB\n"),(int)(0.000512*cardSize));


	if (sd.vol()->fatType() == 0) {
		my_printer.message( T(MSG_SD_Vol_Init_Fail));
		return;
	}
	else
	{
		my_printer.f_message( T(MSG_SD_Vol_Type),(int)sd.vol()->fatType());
	}

	my_printer.echo(T(MSG_SD_Card_OK));

	LsDive (sd.vwd(),0,1);
	sd.vwd()->getName(&current_path[0],70);
	Status = Card_OK;
}

void CardReader::setroot()
{
	sd.chdir(true);
}

void CardReader::release()
{
Status = No_Card;
}

void CardReader::startFileprint()
{
  if(Status == Card_OK || Status == Card_paused)
  {
	  Status = Card_printing;
	  my_printer.print_job_timer.start();
  }
}

void CardReader::pauseSDPrint()
{
	if(Status == Card_printing)
	{
		Status = Card_paused;
		my_printer.print_job_timer.pause();
	}
}


void CardReader::openLogFile(char* name)
{
  logging = true;
  openFile(name, false);
}

void CardReader::getAbsFilename(char *t , uint8_t nb)
{
	if (in_file)
		if (in_file->isOpen())
		{
			in_file->getName(t,nb);
			return;
		}
	*t = '\0';
}

void CardReader::openFile(char* name,bool read, bool replace_current/*=true*/)
{
	if(Status == No_Card)
		return;
	if (in_file && read)
	{
		if(in_file->isOpen())  //replacing current file by new file, or subfile call
		{
//			if(!replace_current)
//			{	// call subfile
//				if((int)file_subcall_ctr>(int)SD_PROCEDURE_DEPTH-1)
//				{
//					my_printer.error();
//					my_printer.f_message(T(MSG_SD_Error_Sublevel),SD_PROCEDURE_DEPTH);
//					kill();
//					return;
//				}
//
//				my_printer.echo();
//				my_printer.f_message(T(MSG_SD_Subroutine),name,&current_file,sdpos);
////				SERIAL_ECHOPGM("SUBROUTINE CALL target:\"");
////				SERIAL_ECHO(name);
////				SERIAL_ECHOPGM("\" parent:\"");
////
////				//store current filename and position
////				//   getAbsFilename(filenames[file_subcall_ctr,13]);
////
////				SERIAL_ECHO(filenames[file_subcall_ctr]);
////				SERIAL_ECHOPGM("\" pos");
////				SERIAL_ECHOLN(sdpos);
//
//				file_subcall[file_subcall_ctr++]=in_file;
//			}
//			else
			{
				in_file->close();
				my_printer.f_message(T(MSG_SD_Printing),name);
				delete in_file;
			}
			in_file = NULL;
		}
	}
	else if (read)//opening fresh file
	{
		file_subcall_ctr=0; //resetting procedure depth in case user cancels print while in procedure
		my_printer.echo();

		my_printer.f_message(T(MSG_SD_Now_Printing),name);
	}
	Status  = Card_OK;


	if (sd.exists(name) && read)
	{
		in_file = new FatFile;
		if (in_file->open(name, O_READ))
		{
			filesize = in_file->fileSize();
			my_printer.f_message(T(MSG_SD_File_Opened),name,filesize);
			sdpos = 0;

			my_printer.f_message(T(MSG_SD_File_Selected));
			//    lcd_setstatus(fname);
		}
		else
		{
			my_printer.f_message(T(MSG_SD_Open_File_Fail),name);
			delete in_file;
			in_file = NULL;
		}
	}
	else
	{
		my_printer.f_message(T(MSG_SD_Open_File_Fail),name);
	}


	if (!read)
	{ //write
		out_file = new FatFile;
		if (!out_file->open(name, O_CREAT | O_APPEND | O_WRITE | O_TRUNC))
		{
			my_printer.f_message(T(MSG_SD_Open_File_Fail),name);
			delete out_file;
			out_file = NULL;
		}
		else
		{
			saving = true;
			my_printer.f_message(T(MSG_SD_Writing),name);
			//     lcd_setstatus(fname);
		}
	}
}

void CardReader::removeFile(char* name)
{
	FatFile file;
	if(Status == No_Card)
		return;
	Status = Card_OK;
	char *filename;

	if (name[0] == '/')
	{	// already an absolute file path
		filename = sf_message(PSTR("%s"),name);
	}
	else
	{	// relative file name, add working directory
		filename = sf_message(PSTR("%s%s"),getPath(),name);
	}

	if (sd.exists(filename))
	{
		if ( FatFile::remove(sd.vwd(),filename))
		{
			my_printer.f_message(T(MSG_SD_File_deleted),filename);
			sdpos = 0;
		}
		else
		{
			my_printer.f_message(T(MSG_SD_Deletion_Error),filename);
		}
	}
}

void CardReader::getStatus()
{
  if(Status == No_Card)
  {
	  my_printer.message(T(MSG_SD_Not_Printing));
   }
  else
  {
	  my_printer.f_message(T(MSG_SD_Printing_Byte),sdpos,filesize);
  }
}

void CardReader::write_command(char *buf)
{
  char* begin = buf;
  char* npos = 0;
  char* end = buf + strlen(buf) - 1;


  if((npos = strchr(buf, 'N')) != NULL)
  {
    begin = strchr(npos, ' ') + 1;
    end = strchr(npos, '*') - 1;
  }
  end[1] = '\r';
  end[2] = '\n';
  end[3] = '\0';
//  file.write(begin);
//  if (file.getWriteError())
//  {
//	  my_printer.error();
//	  my_printer.error_ln(PSTR(MSG_SD_ERR_WRITE_TO_FILE));
//  }
}


void CardReader::checkautostart(bool force)
{
//  if(!force)
//  {
//    if(!autostart_stilltocheck)
//      return;
//    if(autostart_atmillis<millis())
//      return;
//  }
//  autostart_stilltocheck=false;
//  if(Status == No_Card)
//  {
//    initsd();
//    if(Status == No_Card) //fail
//      return;
//  }
//
//  char autoname[30];
//  sprintf_P(autoname, PSTR("auto%i.g"), lastnr);
//  for(int8_t i=0;i<(int8_t)strlen(autoname);i++)
//    autoname[i]=tolower(autoname[i]);
//  dir_t p;
//
//  root.rewind();
//
//  bool found=false;
//  while (root.readDir(&p) > 0)
//  {
//    for(int8_t i=0;i<(int8_t)strlen((char*)p.name);i++)
//    p.name[i]=tolower(p.name[i]);
//    //Serial.print((char*)p.name);
//    //Serial.print(" ");
//    //Serial.println(autoname);
//    if(p.name[9]!='~') //skip safety copies
//    if(strncmp((char*)p.name,autoname,5)==0)
//    {
//      char cmd[30];
//
//      sprintf_P(cmd, PSTR("M23 %s"), autoname);
//      enquecommand(cmd);
//      enquecommand_P(PSTR("M24"));
//      found=true;
//    }
//  }
//  if(!found)
//    lastnr=-1;
//  else
//    lastnr++;
}

void CardReader::closefile(bool store_location)
{
	// closing the input file only
  in_file->sync();
  in_file->close();
  delete in_file;
  in_file = NULL;
  
  if(store_location)
  {
    //future: store printer state, filename and position for continuing a stopped print
    // so one can unplug the printer and continue printing the next day.
    
  }

  
}

void CardReader::getfilename(const uint8_t nr)
{
	if (in_file == NULL)
		in_file = new FatFile;
	if (nr < getnrfilenames())
	{
		in_file->close();
		nrFiles=nr;
		sd.vwd()->rewind();
		while (in_file->openNext(sd.vwd(), O_READ) )
		{
			if (!in_file->isHidden())
			{
				nrFiles--;
			}
			if (nrFiles < 0)
				return;
			in_file->close();
		}
	}
}

uint16_t CardReader::getnrfilenames()
{
	FatFile file;
	nrFiles=0;
	sd.vwd()->rewind();
	while (file.openNext(sd.vwd(), O_READ))
	{
		if (!file.isHidden())
		{
			nrFiles++;
		}
		file.close();
	}
	return nrFiles;
}

void CardReader::chdir(const char * relpath)
{
	char *dirpos = &current_path[0];

	sd.chdir(relpath,true);
	while (*dirpos++) {}; // find end of directory name
	sd.vwd()->getName(dirpos-1,70);
	while (*dirpos++) {}; // find end of new path
	*(dirpos-1) = '/';
	*dirpos = '\0';
}

void CardReader::updir()
{
	char *dirpos = &current_path[0];
  if(current_path[1] =='\0')
	  return;		// already root dir
	while (*dirpos++) {}; // find end of new path
	dirpos = dirpos-2; // the last backslach of the path
	while (*(--dirpos) != '/'){};
	*(dirpos+1) = '\0';
	sd.chdir(current_path,true);
 }


void CardReader::printingHasFinished()
{
	motor.Synchronize();
	if(file_subcall_ctr>0) //heading up to a parent file that called current as a procedure.
	{
		in_file->close();
		delete in_file;
		in_file = file_subcall[file_subcall_ctr--];
		startFileprint();
	}
	else
	{
		motor.QuickStop();
		in_file->close();
		delete in_file;
		in_file = NULL;
		if (out_file)
		{
			out_file->close();
			delete out_file;
		}
		Status = Card_OK;
		if(SD_FINISHED_STEPPERRELEASE)
		{
			motor.FinishAndDisable();
			service.Print(Service_SD_Finished_Release);
		}
		heater.autotempShutdown();
		my_printer.print_job_timer.stop();
	}
}

bool CardReader::IsDir()
{
	if (in_file)
		if (in_file->isOpen())
			return in_file->isDir();
	return false;
}

char* CardReader::FileName()
{
	if (in_file)
	if (in_file->isOpen())
	{
	in_file->getName(&current_file[0],60);
	return current_file;
	}

	current_file[0] = '\0';
	return current_file;
}

#endif //SDSUPPORT
