/*
    K8400 firmware modification based on original Version provided by Vellemann.

 Copyright (C) 2017 Holger Höhnemann

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include "parameter.h"
#include "printer.h"
#include "temperature.h"
#include "language.h"

parameter::parameter()
{
	//Load_Failsave();

}

parameter::~parameter()
{
	// TODO Auto-generated destructor stub
}

void parameter::Load_Failsave()
{
	float tmp1[]=DEFAULT_AXIS_STEPS_PER_UNIT;
	float tmp2[]=DEFAULT_MAX_FEEDRATE;
	long tmp3[]=DEFAULT_MAX_ACCELERATION;
	for (short i=0;i<NUM_AXIS;i++)
	{
		axis_steps_per_unit[i]=tmp1[i];
		max_feedrate_mm_s[i]=tmp2[i];
		max_acceleration_units_per_sq_second[i]=tmp3[i];
	}

	// steps per sq second need to be updated to agree with the units per sq second
	//reset_acceleration_rates();
	uint16_t Checksum_Ckeck_Bytes;

	acceleration=DEFAULT_ACCELERATION;
	retract_acceleration=DEFAULT_RETRACT_ACCELERATION;
	travel_acceleration = DEFAULT_TRAVEL_ACCELERATION;
	min_feedrate_mm_s=DEFAULT_MINIMUMFEEDRATE;
	minsegmenttime=DEFAULT_MINSEGMENTTIME;
	min_travel_feedrate_mm_s=DEFAULT_MINTRAVELFEEDRATE;
	max_jerk[X_AXIS]=max_jerk[Y_AXIS]=DEFAULT_XYJERK;
	max_jerk[Z_AXIS]=DEFAULT_ZJERK;
	max_jerk[E_AXIS]=DEFAULT_EJERK;
	add_homeing[0] = add_homeing[1] = add_homeing[2] = 0;
#ifdef DELTA
	endstop_adj[0] = endstop_adj[1] = endstop_adj[2] = 0;
	delta_radius= DELTA_RADIUS;
	delta_diagonal_rod= DELTA_DIAGONAL_ROD;
	delta_segments_per_second= DELTA_SEGMENTS_PER_SECOND;
	recalc_delta_settings(delta_radius, delta_diagonal_rod);
#endif
#ifdef ULTIPANEL
	plaPreheatHotendTemp = PLA_PREHEAT_HOTEND_TEMP;
	plaPreheatHPBTemp = PLA_PREHEAT_HPB_TEMP;
	plaPreheatFanSpeed = PLA_PREHEAT_FAN_SPEED;
	absPreheatHotendTemp = ABS_PREHEAT_HOTEND_TEMP;
	absPreheatHPBTemp = ABS_PREHEAT_HPB_TEMP;
	absPreheatFanSpeed = ABS_PREHEAT_FAN_SPEED;
#endif
#ifdef ENABLE_AUTO_BED_LEVELING
	zprobe_zoffset = -Z_PROBE_OFFSET_FROM_EXTRUDER;
#endif
#ifdef DOGLCD
	lcd_contrast = DEFAULT_LCD_CONTRAST;
#endif

#ifdef PIDTEMP
#ifdef PID_PARAMS_PER_EXTRUDER
	for (uint8_t e = 0; e < EXTRUDERS; e++)
#else
		int e = 0; UNUSED(e); // only need to write once
#endif
	{
		Kp [e] = DEFAULT_Kp;
		Set_Ki(DEFAULT_Ki,e);
		Set_Kd(DEFAULT_Kd,e);
#ifdef PID_ADD_EXTRUSION_RATE
		Kc [e] = DEFAULT_Kc;
		Kf [e] = DEFAULT_KF;
#endif
	}
	// Extruder offset

	{
	float temp_xoffs[] = EXTRUDER_OFFSET_X;
	float temp_yoffs[] = EXTRUDER_OFFSET_Y;

	for (int ext = 0; ext < EXTRUDERS; ext++)
	{
#if EXTRUDERS > 1
	extruder_offset[X_AXIS][ext] = temp_xoffs[ext];
	extruder_offset[Y_AXIS][ext] = temp_yoffs[ext];
#endif
#ifdef PRESSURE_ADVANCE
	 advance_factor[ext] = 0;							// the pressure advance factor for the extruders
	 advance_delay[ext] = 0;								// the ramp up delay
#endif
	}

#ifdef PIDTEMPBED

     bedKp = DEFAULT_bedKp;
		Set_Ki(DEFAULT_bedKi,HEAT_BED);
		Set_Kd(DEFAULT_bedKd,HEAT_BED);
#endif

		language = lang_en;
	};

	// call updatePID (simil..updatePID();
	heater.UpdatePID();
#endif // PIDTEMP

	Reset_AccelerationRates();

//	SERIAL_ECHO_START;
	my_printer.echo(T(MSG_Default_Parameters_Loaded));

}


void parameter::Load_EEPROM()
{
	int i = EEPROM_OFFSET;
	_EEPROM_readData (i, (unsigned char *)this, sizeof (class parameter));
	i = sizeof (class parameter);
	my_printer.f_message(T(MSG_NB_EEPROM_Bytes_Loaded),i);
	if (Checksum_Error())
	{
		my_printer.f_message(T(MSG_CS_Error_Load_Failsave));
		Load_Failsave();
		Store_EEPROM();
	}

	Reset_AccelerationRates();
	heater.UpdatePID();
}

void parameter::Store_EEPROM()
{
	int i = EEPROM_OFFSET;
	_EEPROM_writeData (i, (unsigned char *)this, sizeof (class parameter));
	Write_Checksum_Bytes();
//	my_printer.f_message(PSTR("cs-offset %i \n"), (int)((long)&checksum_bytes - (long)this) );
//	my_printer.f_message(PSTR("float-offset %i \n"), (int)((long)&axis_steps_per_unit[0] - (long)this) );
}


void parameter::_EEPROM_writeData(int &pos, uint8_t* value, uint8_t size)
{
	do
	{
		eeprom_write_byte((unsigned char*)pos, *value);
		pos++;
		value++;
	}while(--size);
}

void parameter::Reset_AccelerationRates()
{
	for(int8_t i=0; i < NUM_AXIS; i++)
        {
		axis_steps_per_sqr_second[i] = max_acceleration_units_per_sq_second[i] * axis_steps_per_unit[i];
        }
}

void parameter::_EEPROM_readData(int &pos, uint8_t* value, uint8_t size)
{
	do
	{
		*value = eeprom_read_byte((unsigned char*)pos);
		pos++;
		value++;
	}while(--size);
}


void parameter::Print_Settings()
{
#ifndef DISABLE_M503
// Always have this function, even with EEPROM_SETTINGS disabled, the current values will be shown
//	my_printer.echo();
//	SERIAL_ECHOLNPGM("Steps per unit:");
//	my_printer.echo();
//	SERIAL_ECHOPAIR("  M92 X",axis_steps_per_unit[0]);
//	SERIAL_ECHOPAIR(" Y",axis_steps_per_unit[1]);
//	SERIAL_ECHOPAIR(" Z",axis_steps_per_unit[2]);
//	SERIAL_ECHOPAIR(" E",axis_steps_per_unit[3]);
//	SERIAL_ECHOLN("");
	my_printer.message(T(MSG_Steps_Per_Unit));
	my_printer.f_message(T(MSG_NB_Steps_Per_Unit),
			axis_steps_per_unit[0],axis_steps_per_unit[1],axis_steps_per_unit[2],axis_steps_per_unit[3]);


	my_printer.echo();
	my_printer.message(T(MSG_Max_Feedrates));

	my_printer.f_message(T(MSG_NB_Max_Feedrates),
		max_feedrate_mm_s[0], max_feedrate_mm_s[1], max_feedrate_mm_s[2], max_feedrate_mm_s[3]	);


	my_printer.echo();
	my_printer.message(T(MSG_Max_Acceleration));

	my_printer.f_message(T(MSG_NB_Max_Acceleration),
			max_acceleration_units_per_sq_second[0], max_acceleration_units_per_sq_second[1], max_acceleration_units_per_sq_second[2], max_acceleration_units_per_sq_second[3]);

	my_printer.echo();
	SERIAL_ECHOLNPGM("Acceleration: S=acceleration, T=retract acceleration");
	my_printer.echo();
	SERIAL_ECHOPAIR("  M204 S",acceleration );
	SERIAL_ECHOPAIR(" T" ,retract_acceleration);
	SERIAL_ECHOLN("");

	my_printer.echo();
	SERIAL_ECHOLNPGM("Advanced variables: S=Min feedrate (mm/s), T=Min travel feedrate (mm/s), B=minimum segment time (ms), X=maximum XY jerk (mm/s),  Z=maximum Z jerk (mm/s),  E=maximum E jerk (mm/s)");
	my_printer.echo();
	SERIAL_ECHOPAIR("  M205 S",min_feedrate_mm_s );
	SERIAL_ECHOPAIR(" T" ,min_travel_feedrate_mm_s );
	SERIAL_ECHOPAIR(" B" ,minsegmenttime );
	SERIAL_ECHOPAIR(" X" ,max_jerk[X_AXIS] );
	SERIAL_ECHOPAIR(" Z" ,max_jerk[Z_AXIS]);
	SERIAL_ECHOPAIR(" E" ,max_jerk[E_AXIS]);
	SERIAL_ECHOLN("");

	my_printer.echo();
	SERIAL_ECHOLNPGM("Home offset (mm):");
	my_printer.echo();
	SERIAL_ECHOPAIR(" X",add_homeing[0] );
	SERIAL_ECHOPAIR(" Y" ,add_homeing[1] );
	SERIAL_ECHOPAIR(" Z" ,add_homeing[2] );
	SERIAL_ECHOLN("");
#ifdef DELTA
	SERIAL_ECHO_START;
	SERIAL_ECHOLNPGM("Endstop adjustement (mm):");
	SERIAL_ECHO_START;
	SERIAL_ECHOPAIR("  M666 X",endstop_adj[0] );
	SERIAL_ECHOPAIR(" Y" ,endstop_adj[1] );
	SERIAL_ECHOPAIR(" Z" ,endstop_adj[2] );
	SERIAL_ECHOLN("");
	SERIAL_ECHO_START;
	SERIAL_ECHOLNPGM("Delta settings: L=delta_diagonal_rod, R=delta_radius, S=delta_segments_per_second");
	SERIAL_ECHO_START;
	SERIAL_ECHOPAIR("  M665 L",delta_diagonal_rod );
	SERIAL_ECHOPAIR(" R" ,delta_radius );
	SERIAL_ECHOPAIR(" S" ,delta_segments_per_second );
	SERIAL_ECHOLN("");
#endif

#ifdef PIDTEMP
	for (int i = 0; i < EXTRUDERS; i++) {
		my_printer.echo();
		SERIAL_ECHOPAIR("  M301 E", i);
		SERIAL_ECHOPAIR(" P", config.Get_Kp(i));
		SERIAL_ECHOPAIR(" I", config.Get_Ki(i));
		SERIAL_ECHOPAIR(" D", config.Get_Kd(i));
    #ifdef PID_ADD_EXTRUSION_RATE
		SERIAL_ECHOPAIR(" C", config.Get_Kc(i));
		SERIAL_ECHOPAIR(" F", config.Get_Kf(i));
		if (i == 0) SERIAL_ECHOPAIR(" L", lpq_len);
    #endif  // PID_ADD_EXTRUSION_RATE
		SERIAL_ECHOLN("");
	}

#endif // PIDTEMP
#endif
}

void parameter::Set_Ki(float value, int e)
{
	if (e < 0)
		bedKi = value*PID_dT;
	else
		Ki[e] =	value*PID_dT;
    heater.UpdatePID();
}

void parameter::Set_Kd(float value, int e)
{
	if (e<0)
		bedKd = value /PID_dT;
	else
		Kd[e] =	value/PID_dT;
    heater.UpdatePID();
}

float parameter::Get_Ki(int e)
{
	if (e<0)
		return bedKi/PID_dT;
	else
		return Ki[e]/PID_dT;
}

float parameter::Get_Kd(int e)
{
	if (e<0)
		return bedKd*PID_dT;
	else
		return Kd[e]*PID_dT;
}

void parameter::Set_Kp(float value, int e)
{
	if (e < 0)
		bedKp = value;
	else
		Kp[e] =	value;
}

void parameter::Set_Kc(float value, int e)
{
	if (e < 0)
		{} // do nothing
	else
		Kc[e] =	value*PID_dT;
}

void parameter::Set_Kf(float value, int e)
{
	if (e < 0)
		{} // do nothing
	else
		Kf[e] =	value*PID_dT;
}

float parameter::Get_Kp(int e)
{
	if (e<0)
		return bedKp;
	else
		return Kp[e];
}

float parameter::Get_Kc(int e)
{
	if (e<0)
		return 0;
	else
		return Kc[e];
}

float parameter::Get_Kf(int e)
{
	if (e<0)
		return 0;
	else
		return Kf[e];
}

uint16_t parameter::Generate_Check_Sum()
{
// An implementation of the "Fletcher 16" check sum generator

	        uint16_t sum1 = 0xff, sum2 = 0xff;
	        int offset = (int)((long)&checksum_bytes - (long)this);
	        size_t tlen, bytes = sizeof (parameter)-2-offset;
	        uint8_t const * data = (unsigned char *)this +offset+2;

	        while (bytes) {
	                tlen = ((bytes >= 20) ? 20 : bytes);
	                bytes -= tlen;
	                do {
	                        sum2 += sum1 += *data++;
	                        tlen--;
	                } while (tlen);
	                sum1 = (sum1 & 0xff) + (sum1 >> 8);
	                sum2 = (sum2 & 0xff) + (sum2 >> 8);
	        }
	        /* Second reduction step to reduce sums to 8 bits */
	        sum1 = (sum1 & 0xff) + (sum1 >> 8);
	        sum2 = (sum2 & 0xff) + (sum2 >> 8);
	    	my_printer.f_message(T(MSG_CS_Value),((sum2 << 8) | sum1));
	        return (sum2 << 8) | sum1;

}

void parameter::Write_Checksum_Bytes()
{
    uint16_t csum = Generate_Check_Sum();
	int i = EEPROM_OFFSET+(int)((long)&checksum_bytes - (long)this);
	eeprom_write_word((uint16_t *)i, csum);
}

bool parameter::Checksum_Error()
{
	uint16_t csum = Generate_Check_Sum();
	int i = EEPROM_OFFSET+(int)((long)&checksum_bytes - (long)this);
    uint16_t cs_ref = eeprom_read_word((uint16_t *)i);

	return (csum - cs_ref) != 0;
}
