/*
    K8400 firmware modification based on original Version provided by Vellemann.

 Copyright (C) 2017 Holger Höhnemann

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#ifndef temperature_h
#define temperature_h 

//#include <planner_old.h>
#include "MarlinHOH.h"
#ifdef PID_ADD_EXTRUSION_RATE
#include "stepper.h"
#endif
#include "parameter.h"
#include "stopwatch.h"
#include "printer.h"



ISR (TIMER0_COMPB_vect);

ISR (ADC_vect);

class temperatur_Control
{
	friend void __vector_22();
	friend void __vector_29();

public:
	temperatur_Control();
	virtual ~temperatur_Control();
	enum heater_error_numbers {No_Error = 0,MinTempBed, MaxTempBed, MinTemp0, MaxTemp0, MinTemp1, MaxTemp1, BedHeaterBroken, Heater0Broken, Heater1Broken, Heater0Overload, Heater1Overload };

	void Init();
	void Manage();
	void autotempShutdown();
	void Disable();
	void UpdatePID();
	void PID_autotune(float temp, int extruder, int ncycles, bool update );
	bool Tune_Extruder_Fan (int selected_extruder , bool update);

	void Print_Heater_States();
	int Get_Power(int extruder);
	// class variables
	float current_temperature[EXTRUDERS] = ARRAY_BY_EXTRUDERS1 (0.0);
	int target_temperature[EXTRUDERS] = ARRAY_BY_EXTRUDERS1(0);
	float current_temperature_bed = 0.0;
	int target_temperature_bed = 0;

	inline void __attribute__((always_inline)) setTargetHeater(const float &celsius, int extruder) { ((extruder<0) ? target_temperature_bed : target_temperature[extruder]) = celsius; my_printer.Status = Heating;};
//	inline void __attribute__((always_inline)) setHotend(const float &celsius, uint8_t extruder) { current_temperature[extruder] = celsius;	};

	inline float __attribute__((always_inline)) degHotend(int extruder) {return (extruder < 0) ?  current_temperature_bed : current_temperature[extruder];};
	inline float __attribute__((always_inline)) degTargetHotend(int extruder) {return  (extruder < 0) ? target_temperature_bed : target_temperature[extruder];	};
	inline bool __attribute__((always_inline)) isHeatingHotend(int extruder){ return target_temperature[extruder] > current_temperature[extruder];};
	inline bool __attribute__((always_inline)) isCoolingHotend(int extruder) {return target_temperature[extruder] < current_temperature[extruder];};


//	inline void __attribute__((always_inline)) setTargetBed(const float &celsius) {target_temperature_bed = celsius;	};
//	inline float __attribute__((always_inline)) degBed() { return current_temperature_bed;};
//	inline float __attribute__((always_inline)) degTargetBed() {	return target_temperature_bed;	};
	inline bool __attribute__((always_inline)) isCoolingBed() {  return target_temperature_bed < current_temperature_bed;};
	inline bool __attribute__((always_inline)) isHeatingBed() { return target_temperature_bed > current_temperature_bed;	};

	float ADC2Temp(int Extruder);
	void UpdateTemperaturesFromRawValues();
	unsigned char Set_Heater ( int e, int value);
	float Wait_T_Stable(int selected_extruder, bool heater_control = true);
	int Set_FanSpeed (int speed);
	int fanSpeed=0;
	unsigned long Heat (int selected_extruder, unsigned long duration);


#ifdef PREVENT_DANGEROUS_EXTRUDE
  bool allow_cold_extrude = false;
  int extrude_min_temp;
  void set_extrude_min_temp(float temp)
  {
    extrude_min_temp=temp;
  }
  bool tooColdToExtrude(uint8_t e) {
    #if HOTENDS == 1
      UNUSED(e);
    #endif
    return allow_cold_extrude ? false : degHotend(e) < extrude_min_temp;
  }
#else
  static bool tooColdToExtrude(uint8_t e) { UNUSED(e); return false; }
#endif
private:
	class Stopwatch timer;
	int blink_count;

	inline void isr_TIMER0_COMPB_vect();

	inline void isr_ADC_vect();


	void *heater_ttbl_map[EXTRUDERS] = ARRAY_BY_EXTRUDERS( (void *)HEATER_0_TEMPTABLE, (void *)HEATER_1_TEMPTABLE, (void *)HEATER_2_TEMPTABLE );
	uint8_t heater_ttbllen_map[EXTRUDERS] = ARRAY_BY_EXTRUDERS( HEATER_0_TEMPTABLE_LEN, HEATER_1_TEMPTABLE_LEN, HEATER_2_TEMPTABLE_LEN );
#if TEMP_SENSOR_BED > 0
	void *heater_ttbl_map_bed = (void *)BEDTEMPTABLE;
	uint8_t heater_ttbllen_map_bed = BEDTEMPTABLE_LEN;
#endif

	float ADC2Temp(int value,int Extruder);
	//float Wait_T_Stable(int selected_extruder, bool heater_control);

	void Max_Temp_Error(uint8_t e);

	void Min_Temp_Error(uint8_t e);
	//
	// Variablen

	// for ADC-Interrupt
	int ADC_state = -1;
	unsigned long raw_temp_0_value ;
#if EXTRUDERS > 1
	unsigned long raw_temp_1_value ;
#endif
	unsigned long raw_temp_bed_value ;

	// ring buffer for temperature oversampling
	volatile int ADC_Temp0[16];
	volatile int ADC_Temp1[16];
	volatile int ADC_TempBed[16];

	volatile  unsigned int ADC_Lauf;

	// Init min and max temp with extreme values to prevent false errors during startup
	unsigned int minttemp_raw[EXTRUDERS] ;
	unsigned int maxttemp_raw[EXTRUDERS] ;

	int bed_minttemp_raw ;
	int bed_maxttemp_raw ;
	int heater_error = 0;
	int error_count = 0;

	//these variables are for the timer- ISR,
	unsigned char temp_count = 0;
//	unsigned char temp_state = 8;
	volatile bool temp_meas_ready = false;


	// variables for heater control

	unsigned char heater_PWM[EXTRUDERS]  ;
	unsigned char heater_PWM_bed;

	// variables for Heater_measurements
	unsigned long heat[EXTRUDERS];
	unsigned long heat_bed;

	 float temp_iState[EXTRUDERS+1] = ARRAY_BY_EXTRUDERS1(0.0);
	 float temp_dState[EXTRUDERS+1] = ARRAY_BY_EXTRUDERS1(0.0);

	 float temp_iState_min[EXTRUDERS+1];
	 float temp_iState_max[EXTRUDERS+1];
	 float kappa [EXTRUDERS+1];

	float pTerm[EXTRUDERS+1];
	float iTerm[EXTRUDERS+1];
	float dTerm[EXTRUDERS+1];
	float pid_error[EXTRUDERS+1];
	float cTerm[EXTRUDERS+1];
	bool pid_reset[EXTRUDERS+1];

};

extern class temperatur_Control heater;

#endif

